@@ -2,15 +2,14 @@
 # Licensed under the MIT License. See LICENSE in the project root
 # for license information.
 
-from __future__ import absolute_import, division, print_function, unicode_literals
-
+import os
 import psutil
 import pytest
 import sys
 
 import debugpy
 import tests
-from tests import debug, log
+from tests import debug, log, timeline
 from tests.debug import runners
 from tests.patterns import some
 
@@ -42,11 +41,7 @@ def expected_subprocess_config(parent_session):
 
 @pytest.mark.parametrize(
     "start_method",
-    [""]
-    if sys.version_info < (3,)
-    else ["spawn"]
-    if sys.platform == "win32"
-    else ["spawn", "fork"],
+    ["spawn"] if sys.platform == "win32" else ["spawn", "fork"],
 )
 def test_multiprocessing(pyfile, target, run, start_method):
     if start_method == "spawn" and sys.platform != "win32":
@@ -92,7 +87,7 @@ def test_multiprocessing(pyfile, target, run, start_method):
             p.join()
 
         def child(q, a):
-            print("entering child")
+            print("entering child")  # @bp
             assert q.get() == "foo?"
             a.put(Foo())
 
@@ -141,7 +136,20 @@ def test_multiprocessing(pyfile, target, run, start_method):
 
         with debug.Session(child_config) as child_session:
             with child_session.start():
-                pass
+                child_session.set_breakpoints(code_to_debug, all)
+
+            expected_frame = some.dap.frame(code_to_debug, line="bp")
+            stop = child_session.wait_for_stop(
+                "breakpoint",
+                expected_frames=[expected_frame],
+            )
+            child_session.request('stepIn', {"threadId": stop.thread_id})
+
+            stop = child_session.wait_for_stop(
+                "step",
+                expected_frames=[some.dap.frame(code_to_debug, line=expected_frame.items['line'] + 1)],
+            )
+            child_session.request_continue()
 
             expected_grandchild_config = expected_subprocess_config(child_session)
             grandchild_config = child_session.wait_for_next_event("debugpyAttach")
@@ -156,7 +164,8 @@ def test_multiprocessing(pyfile, target, run, start_method):
 
 
 @pytest.mark.parametrize("subProcess", [True, False, None])
-def test_subprocess(pyfile, target, run, subProcess):
+@pytest.mark.parametrize("method", ["startDebugging", "debugpyAttach", ""])
+def test_subprocess(pyfile, target, run, subProcess, method):
     @pyfile
     def child():
         import os
@@ -193,6 +202,8 @@ def test_subprocess(pyfile, target, run, subProcess):
     with debug.Session() as parent_session:
         backchannel = parent_session.open_backchannel()
 
+        if method:
+            parent_session.capabilities["supportsStartDebuggingRequest"] = (method == "startDebugging")
         parent_session.config["preLaunchTask"] = "doSomething"
         parent_session.config["postDebugTask"] = "doSomethingElse"
         if subProcess is not None:
@@ -205,9 +216,19 @@ def test_subprocess(pyfile, target, run, subProcess):
             return
 
         expected_child_config = expected_subprocess_config(parent_session)
-        child_config = parent_session.wait_for_next_event("debugpyAttach")
+
+        if method == "startDebugging":
+            subprocess_request = parent_session.timeline.wait_for_next(timeline.Request("startDebugging"))
+            child_config = subprocess_request.arguments("configuration", dict)
+            del expected_child_config["request"]
+        else:
+            child_config = parent_session.wait_for_next_event("debugpyAttach")
+
+        child_config = dict(child_config)
         child_config.pop("isOutputRedirected", None)
         assert child_config == expected_child_config
+        child_config["request"] = "attach"
+
         parent_session.proceed()
 
         with debug.Session(child_config) as child_session:
@@ -219,7 +240,7 @@ def test_subprocess(pyfile, target, run, subProcess):
             assert str(child_pid) in child_config["name"]
 
             debugpy_file = backchannel.receive()
-            assert debugpy_file == debugpy.__file__
+            assert os.path.abspath(debugpy_file) == os.path.abspath(debugpy.__file__)
 
             child_argv = backchannel.receive()
             assert child_argv == [child, "--arg1", "--arg2", "--arg3"]
@@ -406,8 +427,7 @@ def test_echo_and_shell(pyfile, target, run):
             cwd=os.path.dirname(os.path.abspath(__file__)),
         )
         stdout, _stderr = p.communicate()
-        if sys.version_info[0] >= 3:
-            stdout = stdout.decode("utf-8")
+        stdout = stdout.decode("utf-8")
 
         if "code_to_run.py" not in stdout:
             raise AssertionError(
@@ -466,11 +486,13 @@ def test_subprocess_unobserved(pyfile, run, target, wait):
             if wait:
                 # The child process should not have started running user code until
                 # there was a client connection, so the breakpoint should be hit.
-                child_session.wait_for_stop(expected_frames=[some.dap.frame(child, line="bp")])
+                child_session.wait_for_stop(
+                    expected_frames=[some.dap.frame(child, line="bp")]
+                )
                 child_session.request_continue()
             else:
                 # The breakpoint shouldn't be hit, since that line should have been
-                # executed before we attached. 
+                # executed before we attached.
                 pass
 
             backchannel.send("proceed")
@@ -478,7 +500,9 @@ def test_subprocess_unobserved(pyfile, run, target, wait):
 
 
 @pytest.mark.parametrize("run", runners.all_launch)
-@pytest.mark.skipif(sys.platform != "win32", reason="job objects are specific to Windows")
+@pytest.mark.skipif(
+    sys.platform != "win32", reason="job objects are specific to Windows"
+)
 def test_breakaway_job(pyfile, target, run):
     @pyfile
     def child():
@@ -515,10 +539,12 @@ def test_breakaway_job(pyfile, target, run):
         proc.wait()
 
     with debug.Session() as parent_session:
-        parent_session.config.update({
-            "redirectOutput": False,
-            "subProcess": False,
-        })
+        parent_session.config.update(
+            {
+                "redirectOutput": False,
+                "subProcess": False,
+            }
+        )
         parent_session.expected_exit_code = some.int
         backchannel = parent_session.open_backchannel()
 
@@ -536,4 +562,120 @@ def test_breakaway_job(pyfile, target, run):
         assert backchannel.receive() == "ok"
 
     log.info("Waiting for child process...")
-    child_process.wait()
\ No newline at end of file
+    child_process.wait()
+
+
+@pytest.mark.parametrize("run", runners.all_launch)
+@pytest.mark.skipif(
+    sys.platform == "win32", reason="os.exec() is specific to POSIX"
+)
+def test_subprocess_replace(pyfile, target, run):
+    @pyfile
+    def child():
+        import os
+        import sys
+
+        assert "debugpy" in sys.modules
+
+        from debuggee import backchannel
+
+        backchannel.send(os.getpid())
+
+    @pyfile
+    def parent():
+        import debuggee
+        import os
+        import sys
+
+        debuggee.setup()
+        print(f"execl({sys.executable!r}, {sys.argv[1]!r})")
+        os.execl(sys.executable, sys.executable, sys.argv[1])
+
+    with debug.Session() as parent_session:
+        backchannel = parent_session.open_backchannel()
+        with run(parent_session, target(parent, args=[child])):
+            pass
+
+        expected_child_config = expected_subprocess_config(parent_session)
+        child_config = parent_session.wait_for_next_event("debugpyAttach")
+        child_config.pop("isOutputRedirected", None)
+        assert child_config == expected_child_config
+        parent_session.proceed()
+
+        with debug.Session(child_config) as child_session:
+            with child_session.start():
+                pass
+
+            child_pid = backchannel.receive()
+            assert child_pid == child_config["subProcessId"]
+            assert str(child_pid) in child_config["name"]
+
+
+@pytest.mark.parametrize("run", runners.all_launch)
+def test_subprocess_with_parent_pid(pyfile, target, run):
+    @pyfile
+    def child():
+        import sys
+
+        assert "debugpy" in sys.modules
+
+        import debugpy
+
+        assert debugpy  # @bp
+
+    @pyfile
+    def parent():
+        import debuggee
+        import os
+        import subprocess
+        import sys
+
+        import debugpy
+
+        debuggee.setup()
+
+        # Running it through a shell is necessary to ensure the
+        # --parent-session-pid option is tested and the underlying
+        # Python subprocess can associate with this one's debug session.
+        if sys.platform == "win32":
+            argv = ["cmd.exe", "/c"]
+        else:
+            argv = ["/bin/sh", "-c"]
+
+        cli_opts = debugpy.get_cli_options()
+        assert cli_opts, "No CLI options found"
+
+        host, port = cli_opts.address
+        access_token = cli_opts.adapter_access_token
+
+        shell_args = [
+            sys.executable,
+            "-m",
+            "debugpy",
+            "--connect", f"{host}:{port}",
+            "--parent-session-pid", str(os.getpid()),
+            "--adapter-access-token", access_token,
+            sys.argv[1],
+        ]
+        argv.append(" ".join(shell_args))
+
+        subprocess.check_call(argv, env=os.environ | {"DEBUGPY_RUNNING": "false"})
+
+    with debug.Session() as parent_session:
+        with run(parent_session, target(parent, args=[child])):
+            parent_session.set_breakpoints(child, all)
+
+        with parent_session.wait_for_next_subprocess() as child_session:
+            expected_child_config = expected_subprocess_config(parent_session)
+            child_config = child_session.config
+            child_config.pop("isOutputRedirected", None)
+            assert child_config == expected_child_config
+
+            with child_session.start():
+                child_session.set_breakpoints(child, all)
+
+            child_session.wait_for_stop(
+                "breakpoint",
+                expected_frames=[some.dap.frame(child, line="bp")],
+            )
+            child_session.request_continue()
